from    pprint              import  pprint
from    sqlalchemy.schema   import  Table       , MetaData
from    enum                import  Enum
from    collections         import  OrderedDict

import  saltools.parallel   as      sltp
import  saltools.misc       as      sltm
import  saltools.common     as      sltc
import  saltools.logging    as      sltl

import  csv
import  os

class ExporterType(Enum):
    SQLALCHEMY  = 0
    PPRINT      = 1
    CSV         = 2

class Exporter(sltc.EasyObj):
    EasyObj_PARAMS  = OrderedDict((
        ('type'             , {
                'type'      : ExporterType          ,
                'default'   : ExporterType.PPRINT   ,
            }),
        ('engine_builder'   , {
                'type'      : sltm.SQLAlchemyEBuilder   ,
                'default'   : None                      ,
            }),
        ('path'             , {
                'default'   : None  ,
                'type'      : str   ,
            }),
        ('encoding'         , {
                'default'   : 'utf-8'   ,
                'type'      : str       ,
            }),))
    
    @classmethod
    @sltl.handle_exception  (
        is_log_start= True  ,
        params_start= None  )
    def export_sqlalchemy   (
        cls                     ,
        data                    ,
        engine                  ,
        path        = None      ,
        encoding    = 'utf-8'   ):
        engine  = engine.engine
        def insert_row(
            row                 , 
            table               ,
            meta                , 
            referrer    = None  ,
            referrer_id = None  ):
            columns     = [x for x in row if not isinstance(row[x], list)]
            relations   = [x for x in row if     isinstance(row[x], list)]

            values  = {column: row[column] for column in columns}
            if      referrer != None    :
                values['{}_id'.format(referrer)] = referrer_id
            
            id_     = table.insert().values(**values).execute().lastrowid
            for relation_table_name in relations   :
                relation_table   = Table(relation_table_name, meta, autoload=True)
                for relation_row in row[relation_table_name]   :
                    insert_row(
                        relation_row    ,
                        relation_table  ,
                        meta            ,
                        table.name      ,
                        id_             )

        meta    = MetaData(engine)
        for table_name in data.keys()    :
            table   = Table(table_name, meta, autoload=True)
            for row in data[table_name] :  
                insert_row(row, table, meta)

    @classmethod
    @sltl.handle_exception  (
        is_log_start= True  ,
        params_start= None  )
    def export_pprint       (
        cls                     ,
        data                    ,
        engine      = None      , 
        path        = None      ,
        encoding    = 'utf-8'   ):
        pprint(data)
    
    @classmethod
    @sltl.handle_exception  (
        is_log_start= True  ,
        params_start= None  )
    def export_csv          (
        cls                     ,
        data                    ,
        engine      = None      , 
        path        = None      ,
        encoding    = 'utf-8'   ):
        for key, value in data.items():
            if      len(value) == 0     :
                continue
            if      path == None        :
                path    = f'{key}.csv'
            elif    os.path.isdir(path) :
                path    = os.path.join(path, f'{key}.csv')

            w_headers   = not os.path.isfile(path)
            headers     = value[0].keys()
            with open(path, 'a', newline= '\n', encoding= encoding) as f :  
                writer  = csv.DictWriter(f, fieldnames=headers)
                if      w_headers   :
                    writer.writeheader()
                writer.writerows(value)
    
    def export  (
        self    ,
        data    ):
        export_fn   = getattr(type(self), 'export_'+ self.type.name.lower())
        export_fn(
            data                ,
            self.engine_builder ,
            self.path           ,
            self.encoding       )
        






import  saltools.misc   as      sltm

from    os.path         import  join 

def set_param   (
    param   ,
    value   ):
    globals()[param]    = value
def set_params  (
    root    = './'  ):
    settings_dict   = sltm.g_config(join(root, '__settings.json'))
    _params_list    = {
            'EXTRACTORS_SEPARATOR'  : '-->'     ,
            'KWARGS_SEPARATOR'      : '<->'     ,
            'ESC_OPEN'              : '<('      ,
            'ESC_CLOSE'             : ')>'      ,
            'LIST_SEPARATOR'        : '!!'      ,
            'COLLECTIONS_SEPARATOR' : '|=|'     ,
        }

    for param, def_value in _params_list.items()    :
        set_param(param, settings_dict.get('parameters',{}).get(param, def_value))
    
    set_param('RESOURCE_FOLDER', join(root, 'resources'))

set_params()


'''Scraper object defintion

    Implementation of the scraper object.
'''
from    collections         import  OrderedDict
from    saltools.misc       import  g_path
from    enum                import  Enum

from    .                   import  export          as  slsx
from    .                   import  core            as  slsc
from    .                   import  interface       as  slsi
from    .                   import  extraction      as  slse

import  saltools.logging    as      sltl
import  saltools.parallel   as      sltp

import  queue

class Scraper   (
    sltp.NiceFactory    ):
    EasyObj_PARAMS  = OrderedDict((
            ('start_requests'   , {
                },),
            ('parser'           , {
                'type'  : slsc.Parser   },),
            ('interface'        , {
                'type'      : slsi.Interface    ,
                'default'   : slsi.Requests()   },),
            ('data_exporter'    , {
                'default'   : None          ,
                'type'      : slsx.Exporter },),
            ('requests_adapter' , {
                'default'   : 'REQUEST'         ,
                'type'      : slse.Extractor    },),
        ))

    def _on_stop        (
        self    ,
        factory ):
        self.n_data = 0
    def _on_start       (
        self    ,
        factory ):
        pass
    def _on_init        (
        self    ):
        self.is_no_tasks_stop   = True

        for request in self.start_requests  :
            self.add_request(request)

        self.n_data             = 0
        self.on_stop            = self._on_stop
        self.on_start           = self._on_start
    
    def add_request     (
        self            ,
        request         ):
        request_dict    = {'url': request} if isinstance(request, str) else request
        adapted_req     = self.requests_adapter.extract(
            None            ,
            None            ,
            {}              ,
            **request_dict  )
        if      not isinstance(adapted_req, list)   :
                adapted_req = [adapted_req]

        for request_obj in adapted_req  :
            self.start_tasks.append(
                sltp.FactoryTask(
                    target  = self.execute_request  ,
                    args    = [request_obj]         ))
    @sltl.handle_exception  (
        sltl.Level.ERROR            ,
        is_log_start= True          ,
        params_start= ['request']   ,
        params_exc  = ['request']   )
    def execute_request     (
        self    ,
        request ):
        response    = self.interface.execute_request(request)
        r_list      = []
        data        = self.parser.parse(
            response                        ,
            self.interface.execute_request  ,
            r_list                          ,
            self.data_exporter.export       )
        
        for request in r_list :
            if      request != None :
                self.tasks_queue.put(
                    sltp.FactoryTask(
                        target  = self.execute_request  ,
                        args    = [request]             ))



'''Data extraction logic.
'''
from    saltools.common     import  EasyObj
from    collections         import  OrderedDict
from    saltools.misc       import  g_path      , join_string_array
from    os.path             import  join        , abspath
from    urllib.parse        import  urlencode
from    lxml.html           import  etree
from    enum                import  Enum
from    .                   import  interface   as  slsi
from    .                   import  settings    as  slst

import  saltools.logging    as      sltl
import  saltools.web        as      sltw
import  saltools.misc       as      sltm
import  saltools.common     as      sltc

import  inspect
import  yaml
import  html
import  re

class SourceType    (
    Enum    ):
    '''Source type.
    '''
    REQUEST_URL     = 0
    RESPONSE_URL    = 1
    HTML            = 2
    JSON            = 3
    TEXT            = 4
    CONTENT         = 5
    CONTEXT         = 6

class EXTRACTORS            (
    ):
        SOURCE_TYPE_OBJECT_MAP  = {
            SourceType.REQUEST_URL  : lambda response: response.request_url         ,
            SourceType.RESPONSE_URL : lambda response: response.response_url        ,
            SourceType.HTML         : lambda response: response.html_tree           ,
            SourceType.JSON         : lambda response: response.json                ,
            SourceType.TEXT         : lambda response: response.text                ,
            SourceType.CONTENT      : lambda response: response.content             }
        METHOD_SOURCE_TYPE_MAP  = {
            'XPATH'     : 'HTML'            ,
            'NEXT_PAGE' : 'REQUEST_URL'     ,
            'FROM_JSON' : 'TEXT'            ,
            }
        ABREV_METHOD_MAP        = {
            '='     : 'EQUALS'          ,
            'x'     : 'XPATH'           ,
            'r'     : 'REGEX'           ,
            's'     : 'FROM_HTML'       ,
            'h'     : 'TO_HTML'         ,
            '?'     : 'REPLACE'         ,
            'l'     : 'IN_LIST'         ,
            'j'     : 'FROM_JSON'       ,
            '!'     : 'FILTER'          ,
            'a'     : 'ABS_URL'         ,
            '_'     : 'JOIN_STRS'       ,
            'p'     : 'OBJ_PATH'        ,
            'f'     : 'FORMAT'          ,
            '/'     : 'SLICE'           ,
            'u'     : 'UNESCAPE_HTML'   ,
            'd'     : 'TO_DICT'         ,
            'o'     : 'ARTHM'           ,
            '>'     : 'UPPER'           ,
            '<'     : 'LOWER'           ,
            'rs'    : 'RESOURCE'        ,
            'st'    : 'STRIP'           ,
            ' '     : 'NONE'            ,
            '1'     : 'FIRST'           ,
            'b'     : 'TO_BOOL'         ,

            '@'     : 'REQUEST'         ,
            'n'     : 'NEXT_PAGE'       ,

            'bf'    : 'B_FLATTEN'       ,
            'bm'    : 'B_MULTIPLY'      }

        ABREV_SOURCE_MAP        = {
            'q' : 'REQUEST_URL'     ,
            's' : 'RESPONSE_URL'    ,
            'h' : 'HTML'            ,
            'j' : 'JSON'            ,
            't' : 'TEXT'            ,
            'n' : 'CONTENT'         ,
            'c' : 'CONTEXT'         }
    
        @classmethod
        def g_all_names (
            cls ):
            return [
                x[0] for x in inspect.getmembers(cls, predicate=inspect.ismethod)   \
                if x[0] not in ['g_all_names']]
    #----------------------------------------
    # Data extractors
    #----------------------------------------
        @classmethod
        def EQUALS          (
            cls ,
            r   , 
            c   , 
            x   ,
            y   ):
            '''Check if x == y.
                
                Args:
                    x       (object ): Any `object`.
                    y       (object ): Any `object`.
                Returns:
                    object|None   : x if `x` equals `y` else `None`.
            '''
            return x if x == y else None
        @classmethod
        def XPATH           (
            cls     ,
            r       , 
            c       ,
            x       ,
            xpath   ):
            '''Regex extractor.
                
                Args:
                    x       (object ): The html source or element.
                    xpath   (str    ): Xpath expression.
                Returns:
                    list[object]    : A list of matches.
            '''
            return sltw.g_xpath(x, xpath)
        @classmethod
        def REGEX           (
            cls             ,
            r               ,
            c               ,
            x               ,
            pattern = '.*'  ):
            '''Regex extractor.
                
                Args:
                    x       (str    ): The string to extract from.
                    regex   (str    ): Pattern to match.
                Returns:
                    list[str]   : A list of matches.
            '''
            return re.findall(pattern, x)
        @classmethod
        def FROM_HTML       (
            cls     ,
            r       , 
            c       , 
            x       ):
            '''Gets the HTML source of an HTML element as string.
                
                Args:
                    x   (lxml.html.HtmlElement  ): The element to get the source of.
                Returns:
                    str: HTML source of the element.
            '''
            return etree.tostring(x, encoding='unicode').strip()
        @classmethod
        def TO_HTML         (
            cls     ,
            r       , 
            c       , 
            x       ):
            '''Gets the HTML source of an HTML element as string.
                
                Args:
                    x   (str    ): The html source.
                Returns:
                    lxml.html.HtmlElement   : HTML element.
            '''
            return etree.fromstring(x)
        @classmethod
        def REPLACE         (
            cls                 ,
            r                   ,
            c                   ,
            x                   ,
            pattern     = '\n'  ,
            replacement = ' '   ):
            '''Replaces the pattern with the replacement.

                Args:
                    x           (str)   : The string te process.
                    pattern     (str)   : The regex pattern to replace.
                    replacement (str)   : Replacement string.
                Returns:
                    str : The processed string.
            '''
            return re.sub(pattern, replacement, x)
        @classmethod
        def IN_LIST         (
            cls ,
            r   ,
            c   , 
            x   ):
            '''Put in a list

                Args:
                    x (object)  : The object to put in a list.
                Returns:
                    list[object]    : A list containing the object.
            '''
            return [x]
        @classmethod
        def FROM_JSON       (
            cls ,
            r   , 
            c   , 
            x   ):
            '''Creates a dict from json.

                Args:
                    x   (str    ): JSON string.
                Returns:
                    dict    : A dict object.
            '''
            x   = re.sub(r',\s+,', ',', x)
            return yaml.load(x, yaml.Loader)
        @classmethod
        def OBJ_PATH        (
            cls                     ,
            r                       ,
            c                       ,
            x                       ,
            path            = 0     ,
            is_return_last  = False ,
            is_check_only   = False ):
            '''Gets the value at the given path.

                Args:
                    x               (object                 ): Can be anything from an object, list, or dict.
                    path            (str|int|list[str|int]  ): Can be an integer in case of indices, a string for dict keys
                        or objects attributes, or list containing both, the list can also be a string like the example :
                            - `path_to.0.name` , this will get the value of path_to from the dict or obj, then gets the 
                                element at index 0, then the property/key name.
                    is_return_last  (bool                   ): If `True`, returns the last available value in the path before 
                        failure.
                Returns:
                    object  : The value found at the given path.
            '''
            result  = g_path(x, path, is_return_last= is_return_last)
            if      is_check_only   :
                return True if result != None else False
            else                    :
                return result
        @classmethod
        def FILTER          (
            cls             ,
            r               ,
            c               ,
            x               ,
            path            ,
            value   = None  ):
            '''Filters the values in an iterable.

                Args:
                    x           (collections.abc.Iterable   ): The iterable to filter.
                    path        (str|int|list[str|int]      ): Can be an integer in case of indices, a string for dict keys
                        or objects attributes, or list containing both, the list can also be a string like the example :
                            - `path_to.0.name` , this will get the value of path_to from the dict or obj, then gets the 
                                element at index 0, then the property/key name.
                    value       (object                     ): The value to filter by.
                Returns:
                    list    : A flitered list.
            '''
            return [y for y in x if (
                (value == None and g_path(y, path, is_return_last= False) != None   )\
                or g_path(y,path) == value                                          )]
        @classmethod
        def ABS_URL         (
            cls ,
            r   ,
            c   ,
            x   ):
            '''Gets the absolute url if a url is not absolute.

                Args:
                    x   (str    ): A url.
                Returns:
                    str : Absolute url.
            '''
            return r.host+ x if x[:4] != 'http' else x
        @classmethod
        def JOIN_STRS       (
            cls         ,
            r           ,
            c           ,
            x           ,
            by  = ' '   ):
            '''Joins a list of string.

                Args:
                    x   (list[str]  ): A list of strings.
                    by  (str        ): The string to join by.
                Returns:
                    str : The full string.
            '''
            return join_string_array(x, by)
        @classmethod
        def SLICE           (
            cls         ,
            r           ,
            c           ,
            x           ,
            start   = 0 ,
            end     = -1, 
            step    = 1 ):
            '''Slice a list.

                Args:
                    x       (list   ): The list to slice.
                    start   (int    ): Start index.
                    end     (int    ): End index.
                    step    (int    ): Step.
                Returns:
                    list    : The slice.
            ''' 
            return x[start:end:step]
        @classmethod
        def FORMAT          (
            cls     ,
            r       ,
            c       ,
            x       ,
            f_str   ):
            '''Format a string.

                Args:
                    x       (dictlist|object    ): List or dict of values.
                    f_str   (str                ): Format string.
                Returns:
                    str : The formatted string.
            '''
            x   = x if (isinstance(x,list) or isinstance(x, dict)) else [x]
            return f_str.format(*x)             \
                    if      isinstance(x, list) \
                    else    f_str.format(**x)
        @classmethod
        def UNESCAPE_HTML   (
            cls ,
            r   ,
            c   ,
            x   ):
            '''Unescape HTML.
                Example:
                    - `Kristj&aacuten V&iacute;ctor` >>> `Kristján Víctor`
                Args:
                    x   (str    ): The html to unescape.
                Returns:
                    str : Escaped HTML.
            ''' 
            return html.unescape(x)           
        @classmethod
        def TO_DICT         (
            cls     ,
            r       ,
            c       ,
            x       ,
            keys    ):
            '''Convers a list to a dict.

                Args:
                    x       (list       ): A list.
                    keys    (list[str]  ): A list of keys.
                Reruns:
                    A dict.
            '''
            return {keys[i]: x[i] for i in range(len(keys))}
        @classmethod
        def ARTHM           (
            cls         ,
            r           ,
            c           ,
            x           ,
            y           ,
            op  = '+'   ):
            '''Performs an arithmetic operation.

                Args:
                    y   (float|int  ): Right operand.
                    op  (str        ): Operator +, -, *, **,/, //, %
            '''
            if      op == '+'   :
                return x+ y
            elif    op == '-'   :
                return x- y
            elif    op == '*'   :
                return x* y
            elif    op == '**'  :
                return x** y
            elif    op == '/'   :
                return x/ y
            elif    op == '//'  :
                return x// y
            elif    op == '%'   :
                return x// y
        @classmethod
        def UPPER           (
            cls     ,
            r       ,
            c       ,
            x       ):
            '''Upper case.

                Args:
                    x       (list|dict  ): A string.
                Returns:
                    str :  x in upper case.
            '''
            return x.upper()
        @classmethod
        def LOWER           (
            cls     ,
            r       ,
            c       ,
            x       ):
            '''lower case.

                Args:
                    x       (list|dict  ): A string.
                Returns:
                    str :  x in lower case.
            '''
            return x.lower()
        @classmethod
        def RESOURCE        (
            cls     ,
            r       ,
            c       ,
            x       ):
            '''Loads a resource

                Args:
                    x       (str    ): resource name.
                Returns:
                    dict    :  The dict loaded from the JSON resource.
            '''
            resource_path   = join(slst.RESOURCE_FOLDER, x+'.json')
            return sltm.g_config(resource_path)
        @classmethod
        def STRIP           (
            cls     ,
            r       ,
            c       ,
            x       ):
            '''Strips a string.

                Args:
                    x       (str    ): string.
                Returns:
                    str :  Stripped string.
            '''
            return x.strip()
        @classmethod
        def NONE            (
            cls     ,
            r       ,
            c       ,
            x       ):
            '''Returns the same object.

                Args:
                    x   (object ): An object.
                Returns:
                    object  :  The same object.
            '''
            return x
        @classmethod
        def PRINT           (
            cls         ,
            r           ,
            c           ,
            x           ):
            '''Prints `x` to the console, use for debuging.

                Args:
                    x   (object ): An object.
                Returns:
                    object  :  The same object `x`.
            '''
            print(x)
            return x
        @classmethod
        def FIRST           (
            cls         ,
            r           ,
            c           ,
            x           ):
            '''Returns the element from a list that is not `None`.

                Args:
                    x   (list   ): A list.
                Returns:
                    object  :  The first not `None` object.
            '''
            first   = None
            for e in x :
                if      e not in [None, ''] :
                    first   = e
                    break
            return first
        @classmethod
        def TO_BOOL         (
            cls ,
            r   , 
            c   , 
            x   ):
            '''`True` if x is not `None`, otherwise `False`.
                
                Args:
                    x       (object ): Any `object`.
                Returns:
                    Bool    : True if `x` exists.
            '''
            return True if x != None else False
    #----------------------------------------
    # Buckets extractors
    #----------------------------------------
        @classmethod
        def B_FLATTEN   (
            cls                     ,
            r                       , 
            c                       , 
            x                       , 
            keys                    , 
            is_full_name    = False ):
            '''Flaten nested arrays or dicts.

                Rises all keys elements to the top level:
                    {'b': '1', 'a': {'a1':'1', 'a2':2}} becomes {'b': '1', 'a1':'1', 'a2':2}.
                
                Args:
                    bucket          (list[dict] ): Data bucket.
                    keys            (list[str]  ): The keys to flatten.
                    is_full_name    (bool       ): If `True`, combine sub-bucket value with flattened fields.
                Return:
                    dict    : Flattened bucket.   
            ''' 
            bucket  = x.copy()
            def g_buckets_values(
                b_name      ,
                s_bucket    ):
                values  = {}
                for f_name, field in s_bucket.items() :
                    if      is_full_name   :
                        f_name  = f'{b_name}_{f_name}'
                    values[f_name]   = field
                return values 

            for key in keys :
                if len(bucket[key]) :
                    values  = g_buckets_values(key, bucket[key][0])
                    del bucket[key]
                    bucket.update(values)
            return bucket
        @classmethod
        def B_MULTIPLY  (
            cls                     ,
            r                       ,
            c                       ,
            buckets                 ,
            bucket_A_name           ,
            bucket_B_name           ,
            new_name        = None  ,
            is_del_buckets  = True  ):
            '''Creates a bucket for each combination of bucket A and B.

                Args:
                    buckets         (dict   ): The buckets dict.
                    bucket_A_name   (str    ): Bucket A name.
                    bucket_B_name   (str    ): Bucket B name.
                    new_name        (str    ): The name for the new bucket. If None it is set to `bucket_A_name`.
                    is_del_buckets  (bool   ): If `True`, delete processed buckets.
                Returns:
                    dict    : The new bucket data.
            '''
            buckets     = buckets.copy()
            bucket_1    = buckets[bucket_A_name]   
            bucket_2    = buckets[bucket_B_name]
            if      not new_name    :
                new_name = bucket_A_name

            new_list    = []
            for values_1 in bucket_1 :
                for values_2  in bucket_2 :
                    values_1 = values_1.copy()
                    values_2 = values_2.copy()
                    values_1.update(values_2)
                    new_list.append(values_1)

            if      is_del_buckets :
                del buckets[bucket_A_name]
                del buckets[bucket_B_name]
            
            buckets[new_name]   = new_list
            
            return buckets
    #----------------------------------------
    # Requests extractors
    #----------------------------------------
        @classmethod
        def REQUEST     (
            cls                         ,
            r                           ,
            c                           ,
            x                           ,
            url     = None              ,
            method  = slsi.Method.GET   ,
            params  = {}                ):
            '''Build a request.

                Args:
                    x       (list[str]|list[dict]   ): The job extracted args or a list of urls.
                    url     (str                    ): The url of the request, If `None`, `x` must be a list of urls.
                    method  (interface.Method       ): Request method.
                    params  (dict                   ): The default params. 
                
                Returns:
                    interface.Request   : The genrated request. 
            ''' 
            if      isinstance(method, str) :
                method  = getattr(slsi.Method, method)
            if      not url                 :
                return slsi.Request(x, params= params, method= method)
            else            :
                args_params  = params.copy()
                args_params.update(x)
                return slsi.Request(url, params= args_params, method= method)
        @classmethod
        def NEXT_PAGE   (
            cls             ,
            r               ,
            c               ,
            x               ,
            param   = None  ):
            '''Gets the next page url.

                Args:
                    x       (str    ): If given, it will be used instead of the curnet request url.
                    param   (str    ): If given, it will be parsed first if found.
                Returns:
                    str : The url for the next page. 
            '''
            possible_params = ['page', 'page_number', 'pg', 'p', '']
            page_number     = None
            full_match      = None
            current_url     = x if isinstance(x, str) else r.request_url

            if      param != None   :
                possible_params.insert(0, param)

            for possible_param in possible_params   :
                pattern         = f'{possible_param}'+ r'\w*?[\/=](\d+)'
                pattern_full    = f'{possible_param}'+ r'\w*?[\/=]\d+'
                if      len(re.findall(pattern, current_url)):
                    page_number = int(re.findall(pattern, current_url)[0]) 
                    full_match  = re.findall(pattern_full, current_url)[0]
                    prefix      = full_match[:-len(str(page_number))]
                    break
            
            assert page_number  != None , 'Cannot find a pagination parameter.'

            return re.sub(pattern_full,prefix+str(page_number+1), current_url)
class ExtractorFunction     (
    EasyObj ):
    '''Extractor function.
    '''
    EasyObj_PARAMS  = OrderedDict((
        ('method'       , {
            'default'   : 'JOIN_STRINGS'    ,
            'adapter'   : lambda m          :\
                            getattr(EXTRACTORS, m) if isinstance(m, str) else       \
                            lambda self, r, c, x, **kwargs: m(r, c, x **kwargs)}    ),
        ('kwargs'       , {
            'type'      : dict  ,     
            'default'   : {}    ,}),
        ('is_list'      , {
            'type'      : bool  ,
            'default'   : False }),
        ('source_type'  , {
            'type'      : SourceType            ,
            'default'   : SourceType.CONTEXT    }),))
        
    @classmethod
    def _parse_value        (
        cls     ,
        str_    ):
        '''Parses the value of a kwarg.

            Logic:
                - If the value is raw `<(value)>`, return the value witout further parsing.
                - Try parsing to int, float, bool in order, if all failed, return the value as a string.
            
            Args:
                str_    (str    ): The string to parse.
            Returns:
                object  : The parsed string as an object.
        '''
        open_       = slst.ESC_OPEN.replace('(', r'\(')
        close       = slst.ESC_CLOSE.replace(')', r'\)')
        raw_pattern = f'{open_}(.*?){close}'
        raw_value   = g_path(re.findall(raw_pattern, str_),is_return_last=False)
        if      raw_value != None:
            return raw_value

        types   = [int, float, bool]
        for type_ in types:
            try     :
                return sltc.EasyObj.DEFAULT_PARSERS[type_](str_)
            except  :
                pass
        return str_
    @classmethod
    def _parse_method_type  (
        cls     ,
        str_    ):
        pattern                         = r'^(\*?)(.+?)(?:,(\w+?))?(?::.*)?$$'
        is_list, method_, source_type_  = re.findall(pattern, str_, re.DOTALL)[0]
        
        is_list = True if is_list == '*' else False
        
        if      len(method_) < 4                        :
            method  = EXTRACTORS.ABREV_METHOD_MAP.get(method_)
        else                                            :
            method  = method_
        if      method not in EXTRACTORS.g_all_names()  :
            raise ValueError(f'Method name/abrv {method_} is not found.')
        
        source_type = source_type_
        if      source_type_ == ''                              :
            source_type = 'CONTEXT'
        elif    len(source_type_) < 4                           :
            source_type = EXTRACTORS.ABREV_SOURCE_MAP.get(source_type)
        if      source_type not in SourceType._member_names_    :
            raise ValueError(f'Source type {source_type_} is not found.')
        
        return is_list, method, source_type
    @classmethod
    def _parse_kwargs       (
        cls     ,
        str_    ):
        kwargs  = {}
        is_list, method, source_type    = cls._parse_method_type(str_)
        pattern_kwargs                  = r'(?:.*?):(.*)'
        pattern                         = r'(?:([\w*]*?)?=)?(.+?)(?:<->|$)'
        kwargs_str                      = re.findall(pattern_kwargs, str_)

        if      len(kwargs_str) != 0    :
            kwargs_str                      = kwargs_str[0]
            kwargs_strs                     = re.findall(pattern, kwargs_str)
            names   = getattr(EXTRACTORS, method).__code__.co_varnames[4:]
            try     :
                for i in range(len(kwargs_strs)):
                    k_name  = kwargs_strs[i][0]
                    is_kist = k_name[:1] == '*'
                    k_name  = k_name.replace('*', '')

                    if      k_name == ''    :
                        k_name  = names[i]
                    kwargs[k_name]  = [cls._parse_value(x) for x in kwargs_strs[i][1].split(slst.LIST_SEPARATOR)]   \
                                        if is_kist                                  \
                                        else   cls._parse_value(kwargs_strs[i][1]   )
            except  :
                raise ValueError(f'Kwargs cannot be parsed {str_}')
        return {
            'is_list'       : is_list       , 
            'method'        : method        , 
            'source_type'   : source_type   , 
            'kwargs'        : kwargs        }
    @classmethod
    def _EasyObj_parser     (
        cls     ,
        *args   ,
        **kwargs):
        '''Parsing logic for ExtractorFunction.
        '''
        if      not len(kwargs) and len(args) == 1 and isinstance(args[0], str) :
            return [], cls._parse_kwargs(args[0])
        else                                                                    :
            return args, kwargs

    @sltl.handle_exception  (
        sltl.Level.ERROR    ,
        params_exc          = [
                'self.method'       ,
                'self.kwargs'       ,
                'self.source_type'  ,
                'self.is_list'      ,
                'r'                 ,
            ]               )
    def extract             (
        self        , 
        r           , 
        c           ,
        x           , 
        **kwargs    ):
        '''Extraction logic.
        '''
        kwargs.update(self.kwargs)

        if      self.is_list    :
            return [self.method(r, c, y, **kwargs) for y in x]
        else                    : 
            return self.method(r, c, x, **kwargs)
class ExtractorCollection   (
    EasyObj ):
    EasyObj_PARAMS  = OrderedDict((
        ('functions'    , {
            'type'      : [ExtractorFunction]   ,
            'default'   : ['JOIN_STRS']         },),))
    
    @classmethod
    def _EasyObj_parser     (
        cls     ,
        *args   ,
        **kwargs):
        '''Parsing logic for Extractor.

            Logic :
                - If `functions` is in kwargs, put the value into args, empty kwargs.
                - If the first arg is a string, split by `slst.EXTRACTORS_SEPARATOR` and put into args.
        '''
        kwargs  = kwargs.copy()
        args    = list(args).copy()

        if      kwargs.get('functions')   :
            args    = [kwargs['functions']]
            del kwargs['functions']
        if      isinstance(args[0], str)    :
            args = [args[0].split(slst.EXTRACTORS_SEPARATOR)]
        return args, kwargs

    def extract             (
        self    ,
        r       ,
        c       ,
        x       ,
        **kwargs):
        start_fn    = self.functions[0]
        if      c == None       and \
                r != None           :
            if      start_fn.source_type == SourceType.CONTEXT  :
                source_type_    = EXTRACTORS.METHOD_SOURCE_TYPE_MAP.get(start_fn.method.__name__, 'TEXT')
                source_type_    = getattr(SourceType, source_type_)
            else                                                : 
                source_type_    = start_fn.source_type
            c   = EXTRACTORS.SOURCE_TYPE_OBJECT_MAP[source_type_](r)
        if      x == None       :
            if      start_fn.source_type == SourceType.CONTEXT  :
                x   = c
            else                                            :
                x   = EXTRACTORS.SOURCE_TYPE_OBJECT_MAP[start_fn.source_type](r)

        for function in self.functions:
            x = function.extract(r, c, x, **kwargs)
        return x
class Extractor             (
    EasyObj ):
    '''Data extractor
    '''
    EasyObj_PARAMS  = OrderedDict((
        ('collections'  , {
            'type'      : [ExtractorCollection  ],
            'default'   : ['JOIN_STRS']         },),))
    
    @classmethod
    def _EasyObj_parser     (
        cls     ,
        *args   ,
        **kwargs):
        '''Parsing logic for Extractor.

            Logic :
                - If `collections` is in kwargs, put the value into args, empty kwargs.
                - If the first arg is a string, split by `slst.COLLECTIONS_SEPARATOR` and put into args.
        '''
        kwargs  = kwargs.copy()
        args    = list(args).copy()

        if      kwargs.get('collections')   :
            args    = [kwargs['collections']]
            del kwargs['collections']
        if      isinstance(args[0], str)    :
            args = [args[0].split(slst.COLLECTIONS_SEPARATOR)]
        return args, kwargs

    @sltl.handle_exception  (
        sltl.Level.ERROR            ,
        params_exc          = None  )
    def extract             (
        self    ,
        r       ,
        c       ,
        x       ,
        **kwargs):
        values  = []
        for collection in self.collections  :
            values.append(collection.extract(r, c, x, **kwargs))
        return values if len(self.collections) > 1 else values[0]



import  argparse
import  os 
from   .            import  project    as slsp

def parse_args  (
    ):
    parser  = argparse.ArgumentParser('salscraper')
    parser.add_argument(
        'path'                                                  , 
        type    = str                                           ,
        help    = 'Path to a project folder or a scraper file.' )
    parser.add_argument(
        '-u'                        ,
        '--url'                     ,
        type    = str               ,
        help    = 'Url to scrape.'  )
    return vars(parser.parse_args())

def main        (
    ):
    args_dict   = parse_args()
    path        = args_dict['path']
    url         = args_dict.get('url')
    if      os.path.isfile(path)   :
        slsp.Project.run_scraper(path, url)
    elif    os.path.isdir(path)    :
        p   = slsp.Project(
            n_workers   = 5     ,
            root_dir    = path  )
        p.start()
        p.join_exit()

main()


from    decimal             import  Decimal
from    collections         import  OrderedDict
from    collections.abc     import  Iterable    , Callable
from    enum                import  Enum

from    .                   import  interface   as slsi
from    .                   import  extraction  as slse
from    dateutil.parser     import  parse       as dparse
from    datetime            import  datetime    as dt

import  saltools.logging    as      sltl
import  saltools.common     as      sltc
import  saltools.misc       as      sltm

import  re

FIELD_NONE  = 'Just a string to represent a null value'

class FieldType (
    Enum    ):
    '''Field type.
    '''
    INTEGER             = 0
    FLOAT               = 1
    DECIMAL             = 3
    STRING              = 4
    BOOL                = 5
    DATETIME            = 6
    DATETIME_STR        = 7
    DATETIME_STR_END    = 8
    REQUEST             = 9

class ContainerBase (
    sltc.EasyObj    ):
    '''Base for data containers/processors.

        Args:
            id_         (str                        ): An id.
            extractor   (slse.Extractor             ): A single `Extractor` or a list.
            adapter     (collections.abc.Callable   ): Additional adaption logic for the value returned 
                by the extractor.
    '''
    EasyObj_PARAMS  = OrderedDict((
        ('extractor'    , {
            'type'   : slse.Extractor   ,
            'default': None             }),
        ('id_'          , {
            'type'      : str   ,
            'default'   : 'N/A' }),))
    
    def _g_noxt_next    (
        cls         ,
        requests    ):
        requests_noxt   = []
        requests_next   = []

        for request in requests :
            methods = [func.method for coll in request.extractor.collections for func in coll.functions]
            if      slse.EXTRACTORS.NEXT_PAGE in methods    :
                requests_next.append(request)
            else                                            :
                requests_noxt.append(request)
        return requests_noxt, requests_next

    def _extract    (
        self        ,
        r           ,
        c           ):
        '''Generic extraction logic.

            Uses the extractor to get the data from the source.
            Args:
                r       (interface.Response ): Server response.
                c       (object             ): Specific context to be used.
            Returns:
                object  : The parsed value of the field.
        '''
        if      self.extractor != None  :
            return self.extractor.extract(r, c, None)
        else                            :
            return None
    
    def extract     (
        self                ,
        r                   ,
        c           = None  ,
        r_executer  = None  ,
        parser      = None  ,
        r_list      = None  ):
        '''Extracts the data.

            Further manipulation of the data returned by `self._extract`
            Must be overridden 
            Args:
                r   (interface.Response ): Server response.
                c   (object             ): Specific context to be used.
            Returns:
                object  : The parsed value of the field.
        '''
        raise NotImplementedError
class Field         (
    ContainerBase   ):
    EasyObj_PARAMS          = OrderedDict((
        ('type'             , {
            'type'   : FieldType    ,
            'default': 'STRING'     }),
        ('value'            , {
            'default': FIELD_NONE   }),
        ('is_return_value'  , {
            'type'      : bool  ,
            'default'   : False }),
        ('is_save'          , {
            'type'      : bool  ,
            'default'   : True  }),
        ('data_adapter' , {
            'type'      : slse.Extractor    ,
            'default'   : None              }),))
    FIELD_TYPE_OBJECT_MAP   = {
            FieldType.INTEGER   : lambda x: int(float(x))               ,
            FieldType.FLOAT     : float                                 ,
            FieldType.DECIMAL   : Decimal                               ,
            FieldType.STRING    : str                                   ,
            FieldType.BOOL      : sltc.EasyObj.DEFAULT_PARSERS[bool]    ,
            FieldType.DATETIME  : sltc.EasyObj.DEFAULT_PARSERS[dt]      ,
        }
    
    @classmethod
    def _parse_value (
        cls                     ,
        str_                    ,
        type_                   ,
        is_return_value = False ):
        '''Parses the a string using the correct type parsing logic.

            Args:
                str_            (str        ): The string to parse.
                type_           (FieldType  ): Type.
                is_return_value (bool       ): If True, return str_ on failure, could be harmful when saving to
                    a database.
            Return:
                object  : The parsed value.
        '''
        if      str_ in [None, '']  :
            return None
        value   = None
        if      type_ == FieldType.REQUEST          :
            if      not isinstance(str_, list)  :
                str_    = [str_]
            for x in str_   :
                assert isinstance(x, slsi.Request), 'Value returned by extractor is not of type REQUEST'
            value   = str_        
        elif    type_ == FieldType.DATETIME_STR     :
            try     :
                value   = cls.FIELD_TYPE_OBJECT_MAP[FieldType.DATETIME](str_).isoformat()
                value   = value[:19]
            except  :
                value   = str_
        elif    type_ == FieldType.DATETIME_STR_END :
            try     :
                value   = cls.FIELD_TYPE_OBJECT_MAP[FieldType.DATETIME](str_, False).isoformat()
                value   = value[:19]
            except  :
                value   = str_
        else                                        :
            try     :
                value   = cls.FIELD_TYPE_OBJECT_MAP[type_](str_)
            except  :
                if      is_return_value :
                    value   = str_
        return value

    @sltl.handle_exception  (
        sltl.Level.ERROR    ,
        params_exc          = [
                'self.id_'  ,
                'r'
            ]               )
    def extract             (
        self                ,
        r                   ,
        c           = None  ,
        r_executer  = None  ,
        parser      = None  ,
        r_list      = None  ):
        if self.value != FIELD_NONE :
            return self.value

        value   = self._extract(r, c)
        value   = value if self.data_adapter == None else self.data_adapter.extract(r, c, value)
        value   = self._parse_value(value, self.type, self.is_return_value)
        
        if      self.type       != FieldType.REQUEST    \
                or r_executer   == None                 :
            ret_value   = value
        else                                            :
            value = r_executer(value[0])
            value = parser.parse(value, r_executer, r_list)
            ret_value   = list(value.values())[0]
        return ret_value
        
class Bucket        (
    ContainerBase   ):
    EasyObj_PARAMS          = OrderedDict((
        ('fields'       , {
            'type'      : [ContainerBase]   ,
            'default'   : None              }),
        ('is_skip_None' , {
            'type'      : bool  ,
            'default'   : False }),
        ('data_adapter' , {
            'type'      : slse.Extractor    ,
            'default'   : None              }),))
    
    @classmethod
    def _EasyObj_parser     (
        cls     ,
        *args   ,
        **kwargs):
        '''Parsing logic for Bucket.

            Logic:
                - If the fields are in `args`, move into `kwargs['fields']`
                - For each field, check if the field's kwargs/args to determine if a sub bucket:
                    - If `args` and `args[2]` is a list or `kwargs` and there is a key 'field'.
        '''
        args    = list(args).copy()
        kwargs  = kwargs.copy()

        if      len(args) > 2   :
            kwargs['fields'] = args[2]
            del args[2]
        
        for i in range(len(kwargs['fields'])):
            value   = kwargs['fields'][i]
            if      isinstance(value, list) :
                if      isinstance(sltm.g_path(value,2, is_return_last=False), list)    :
                    kwargs['fields'][i] = Bucket(*value)
                else                                                                    :
                    kwargs['fields'][i] = Field(*value)
            elif    isinstance(value, dict) :
                if      'fields' in value   :
                    kwargs['fields'][i] = Bucket(**value)
                else                        :
                    kwargs['fields'][i] = Field(**value)
        
        return args, kwargs
    
    def extract (
        self                ,
        r                   ,
        c           = None  ,
        r_executer  = None  ,
        parser      = None  ,
        r_list      = None  ):
        contexts    = self._extract(r, c) if self.extractor != None else [c]
        if      contexts == None                :
            contexts    = [c]
        if      not isinstance(contexts, list)  :
            contexts    = [contexts]
        data        = []
        for context in contexts :
            if      self.is_skip_None and context == None   :
                continue
            data.append(
                    {   
                        field.id_: field.extract(
                            r           , 
                            context     ,
                            r_executer  ,
                            parser      ,
                            r_list      ) for field in self.fields
                    }
                )
        
        return self.data_adapter.extract(r, c, data) if self.data_adapter != None else data
class Rule          (
    ContainerBase   ):
    EasyObj_PARAMS  = OrderedDict((
        ('buckets'          , {
            'type'      : [Bucket]  ,
            'default'   : []        }),
        ('requests'         , {
            'type'      : [Field]   ,
            'default'   : []        }),
        ('data_adapter'     , {
            'type'      : slse.Extractor    ,
            'default'   : None              }),
        ('is_stop_empty'    , {
            'type'      : bool  ,
            'default'   : True  }),))
    
    def _on_init    (
        self    ):
        for request in self.requests    :
            request.type    = FieldType.REQUEST
    
    def extract (
        self                ,
        r                   ,
        c           = None  ,
        r_executer  = None  ,
        parser      = None  ,
        r_list      = None  ):
        if      self._extract(r, c) in [None, []]   :
            return
        
        data        = {
            b.id_   : b.extract(
                r           ,
                c           ,
                r_executer  ,
                parser      ,
                r_list      ) for b in self.buckets }
        len_data    = sum(map(len, data.values()))
        requests_noxt, requests_next = self._g_noxt_next(self.requests)
        requests    = []
        for request in requests_noxt+ requests_next :
            if      request in requests_next    and \
                    self.is_stop_empty          and \
                    len(requests)   == 0        and \
                    len_data        == 0            :
                    break
            x           = request.extract(r, c)
            if      isinstance(x, list) :
                requests.extend(x)
            else                        :
                requests.append(x)
            
        if      self.data_adapter != None   :
            data = self.data_adapter.extract(r, None, data)
        return data, requests
class Parser        (
    sltc.EasyObj    ):
    EasyObj_PARAMS  = OrderedDict((
        ('rules'        , {
            'type'  : [Rule]    }),))

    @sltl.handle_exception  (
        params_exc  = ['r'] )
    def parse               (
        self                ,
        r                   ,
        r_executer          ,
        r_list              ,
        exporter    = None  ):
        data            = {}
        rule_found      = False
        for rule in self.rules:
            extracted   = rule.extract(
                r          ,
                None       ,
                r_executer , 
                self       ,
                r_list     )
            if      extracted == None   :
                continue
            rule_found  = True

            e_buckets, e_requests   = extracted
            for bucket in e_buckets   :
                data[bucket] = e_buckets[bucket]
        
            for e_request in e_requests : 
                r_list.append(e_request)

        assert rule_found, 'No rule is found for the given response.'
        
        if      exporter and len(data) :
            exporter(data)
        
        return data
                


'''Web interface.

    An interface.
'''
from    saltools            import  logging     as stl

from    enum                import  Enum
from    saltools.web        import  do_request
from    saltools.common     import  EasyObj
from    urllib.parse        import  urlparse
from    collections         import  OrderedDict
from    lxml.html           import  fromstring

import  pickle
import  json

class Method    (
    Enum    ):
    GET     = 0
    POST    = 1
    JSON    = 2
    PUT     = 3

class Request   (
    EasyObj ):
    EasyObj_PARAMS  = OrderedDict((
        ('url'          , {}),
        ('method'       , {
                'type'      : Method        ,
                'default'   : Method.GET    ,
            }),
        ('params'       , {
                'default': {}   ,
            }),
        ('cookies'      , {
                'default': None ,
            }),
        ('headers'      , {
                'default': None ,
            }),
        ('session'      , {
                'default': None ,
            }),))

    def __str__ (
        self    ):
        return '{}: {}'.format(self.url, self.params)
    def __repr__(
        self    ):
        return str(self)

    def _on_init(
        self    ):
        uri         = urlparse(self.url)
        self.host   = f'{uri.scheme}://{uri.netloc}'

        if      self.headers == None    :
            self.headers    = Interface.HEADERS
class Response  (
    EasyObj ):
    ''' A server response.
    '''
    EasyObj_PARAMS  = OrderedDict((
        ('response_url' , {}        ),
        ('request_url'  , {}        ),
        ('request_obj'  , {}        ),
        ('status_code'  , {}        ),
        ('content'      , {}        ),
        ('text'         , {}        ),
        ('is_redirect'  , {}        ),
        ('cookies'      , {}        ),
        ('headers'      , {}        ),
        ('session'      , {}        )))
    
    @classmethod
    def pickle_dump (
        cls     ,
        r       ,
        path    ):
        html_tree   = r.html_tree
        r.html_tree = None

        with open(path, 'wb') as f:
            pickle.dump(r, f)

        r.html_tree = html_tree
    @classmethod
    def pickle_load (
        cls     ,
        path    ):
        with open(path, 'rb') as f:
            r = pickle.load(f) 
        
        try     :
            r.html_tree = fromstring(r.text)
        except  :
            pass
        
        return r
        
    def __str__ (
        self    ):
        return f'{self.request_url}: {self.status_code}'
    def __repr__(
        self    ):
        return str(self)  
    
    def _on_init(
        self    ):
        uri         = urlparse(self.request_url)
        self.host   = f'{uri.scheme}://{uri.netloc}'
        try         :
            self.html_tree  = fromstring(self.text)
        except      :
            self.html_tree  = None 
        try         :
            self.json       = json.loads(self.text)
        except      :
            self.json       = None

class Interface (
    EasyObj ):
    EasyObj_PARAMS  = OrderedDict((
            ('timeout'  , {
                    'default'   : 10.0  ,
                    'type'      : float ,
                }   ),
        ))

    #Default headers
    HEADERS = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
        } 
    
    def execute_request(
        self    , 
        request ):
        pass
class Requests  (
    Interface   ):
    '''Based on requests module.

        Web interface based on the requests module.
    '''

    @stl.handle_exception   (
        params_exc  = [
                'request',
            ]   )
    def execute_request     (
        self    , 
        request ):
        if      isinstance(request, str)        :
            request     = Request(url= request)

        if      request.method == Method.GET    :
            r, session = do_request(
                request.url                 , 
                request.params              , 
                headers = request.headers   ,
                timeout = self.timeout      )
        elif    request.method == Method.POST   :
            r, session = do_request(
                request.url                 , 
                request.params              ,
                is_post = True              ,
                headers = request.headers   ,
                timeout = self.timeout      )
        elif    request.method == Method.JSON   :
            r, session = do_request(
                request.url                 ,
                request.params              ,
                is_json = True              ,
                headers = request.headers   ,
                timeout = self.timeout      )
        return Response(
            response_url= r.url         ,
            request_obj = request       ,
            request_url = request.url   ,
            status_code = r.status_code ,
            content     = r.content     ,
            text        = r.text        ,
            is_redirect = r.is_redirect ,
            cookies     = r.cookies     ,
            headers     = r.headers     ,
            session     = session       )


from    .                   import  export      as  slsx
from    .                   import  extraction  as  slse
from    datetime            import  datetime    as  dt
from    collections         import  OrderedDict
from    .scraper            import  Scraper

import  saltools.logging    as      sltl
import  saltools.parallel   as      sltp
import  saltools.misc       as      sltm
import  saltools.common     as      sltc
import  saltools.files      as      sltf

import  importlib.util
import  string
import  random
import  json
import  os

class Project(
    sltp.NiceFactory    ):
    EasyObj_PARAMS  = OrderedDict((
        ('root_dir' , { }),))
    
    @classmethod
    def _g_json         (
        cls     ,
        path    ):
        name        = os.path.split(path)[-1][:-5]
        json_dict   = sltm.g_config(path)
        scraper     = Scraper(**json_dict['scraper'  ])
        if      json_dict.get('logger') :
            scraper.logger  = sltc.EasyObj.select_type(**json_dict['logger'])
        return scraper, name
    @classmethod
    def _g_py           (
        cls     ,
        path    ):
        module, name    = sltm.load_module(path)
        return  module.g_scraper(), name
    @classmethod
    def _load_customs   (
        cls         ,
        script_path ):
        path    = os.path.join(
            os.path.dirname(script_path)    ,
            '__custom_extractors.py'        )
        if      not os.path.isfile(path) :
            return
        custom_exts, name   = sltm.load_module(path)
        for ext_name in dir(custom_exts.EXTRACTORS):
            if      callable(getattr(custom_exts.EXTRACTORS, ext_name))     and \
                    not ext_name.startswith('__')                               :
                setattr(slse.EXTRACTORS, ext_name, getattr(custom_exts.EXTRACTORS, ext_name))   
    @classmethod
    def _start_scraper  (
        cls                 ,
        name                ,
        scraper             ,
        settings    = {}    ,
        is_join     = True  ):
        def g_default_logger(
            name        ,
            settings    ):
            default_logger_dict = {
                    'type'      : 'ConsoleLogger'   ,
                    'kwargs'    : {}                ,
                }
            default_logger      = settings.get('default_logger', default_logger_dict)
            default_logger      = getattr(sltl, default_logger['type'])(**default_logger['kwargs'])
            prefix              = dt.utcnow().isoformat()
            default_logger.id_  = f'{name}-{prefix}'
            return default_logger
        
        if      scraper.logger        == None   :
            scraper.logger  = g_default_logger(name, settings)
        if      scraper.data_exporter == None   :
            scraper.data_exporter   = slsx.Exporter(**settings['data_exporter'])
        sltl.set_main_logger(scraper.logger)
        scraper.logger.start()
        scraper.start()
        if      is_join:
            scraper.join_exit()
    
    @classmethod
    def run_scraper     (
        cls             ,
        path            ,
        url             ,
        is_join = True  ):
        cls._load_customs(path)
        settings    = sltm.g_config(os.path.join(
            os.path.dirname(path)   ,
            '__settings.json'       ))

        if      path.split('.')[-1] == 'json'   :
            scraper, name   = cls._g_json(path)
        elif    path.split('.')[-1] == 'py'     :
            scraper, name   = cls._g_py(path)
        
        if      url != None :
            scraper.start_tasks.clear()
            scraper.add_request(url)
        
        cls._start_scraper(
            name        ,
            scraper     ,
            settings    ,
            is_join     )
    
    def _on_init        (
        self    ):
        if      not os.path.isdir(self.root_dir):
            os.makedirs(self.root_dir)
        self.id_        = os.path.dirname(self.root_dir)
        
        for task in self._load_scrapers():
            self.start_tasks.append(task)
    def _load_scrapers  (
        self    ):
        paths   = sltf.g_filders(
            self.root_dir               ,
            '^(?!__)\w+\.(?:py|json)$'  ,
            True                        ,
            True                        ,
            False                       ,
            False                       )
        
        paths   = [os.path.abspath(x) for x in paths]
        return  [
                sltp.FactoryTask(
                    id_         = os.path.split(path)[-1]   ,
                    target      = type(self).run_scraper    ,
                    args        = [path, None, False]       ,
                    is_process  = True                      ) for path in paths
            ]
        


    



from    setuptools  import  setup   , find_packages
from    os.path     import  join    , dirname         , abspath
import  sys

BASE_DIR    = abspath(dirname(__file__))
README      = open(join(BASE_DIR, 'README.rst')).read()
REQS        = [line.strip() for line in open(join(BASE_DIR, 'requirements.txt')).readlines() if line.strip()]

setup(
    name                            = 'salscraper'                                  ,
    version                         = '0.2.1'                                       ,
    description                     = 'A scarping tool'                             ,
    long_description                = README                                        ,
    long_description_content_type   = 'text/markdown'                               ,
    classifiers                     = [
          'Development Status :: 4 - Beta'                                          ,
          'Intended Audience :: Developers'                                         ,
          'License :: OSI Approved :: MIT License'                                  ,
          'Operating System :: Microsoft :: Windows'                                ,
          'Programming Language :: Python'                                          ,
          'Topic :: Software Development :: Libraries :: Python Modules'            ],
    keywords                        = 'scrape crawl'                                ,
    author                          = 'saledddar'                                   ,
    author_email                    = 'saledddar@gmail.com'                         ,
    url                             = 'https://github.com/Saledddar/salscraper'     ,
    license                         = 'MIT'                                         ,
    include_package_data            = True                                          ,
    zip_safe                        = False                                         ,
    packages                        = ['salscraper']                                ,
    package_dir                     = {'': 'src'}                                   ,
    install_requires                = [REQS]                                        )
